(* Title:  Specify/pre-conds.sml
   Author: Walther Neuper 110226
   (c) due to copyrigh,t terms
*)

signature PRE_CONDITIONS =
sig
  type T
  type unchecked
(*type unchecked_pos*)
  type unchecked_TEST
  type checked
(*type checked_pos*)
  type checked_TEST

  type env_subst
  type env_eval
  type input

  val to_string : Proof.context -> T -> string

  val max_variant: Model_Def.i_model -> int
  val environment: Model_Def.i_model -> Env.T

  val check: Proof.context -> Rule_Set.T -> unchecked -> Model_Def.i_model -> 'a -> checked
  val check_TEST: Proof.context -> Rule_Def.rule_set -> unchecked_TEST -> env_eval -> checked_TEST
  val check_OLD: Proof.context -> Rule_Def.rule_set -> unchecked ->
    Model_Pattern.T * Model_Def.i_model_TEST -> checked

  val eval: Proof.context -> Rule_Set.T -> bool * term -> bool * term
  val feedback_TEST_to_string: Model_Def.i_model_feedback_TEST -> string

  val check_variant_envs: Proof.context -> Rule_Set.T -> unchecked ->
    env_subst -> env_eval -> checked
  val sep_variants_envs: Model_Pattern.T -> Model_Def.i_model_TEST ->
    ((Model_Def.i_model_TEST * env_subst * env_eval) * Model_Def.variant) list
  val sep_variants_envs_OLD: Model_Pattern.single list -> Model_Def.i_model_TEST ->
    env_subst * env_eval;

(*from isac_test for Minisubpbl*)
  val mk_env: Model_Def.i_model_feedback_TEST -> Env.T
  val mk_env_subst: (Model_Pattern.single * Model_Def.i_model_single_TEST) list -> env_subst
  val mk_env_eval: (Model_Pattern.single * Model_Def.i_model_single_TEST) list -> env_eval
  val arrange_args2: 'a list -> 'b list -> 'c list -> int * 'd list -> (('a * 'b * 'c) * 'd) list
  val get_i_model: (Model_Pattern.single * Model_Def.i_model_single_TEST) list ->
    Model_Def.i_model_TEST
  val filter_variants: (Model_Pattern.single * Model_Def.i_model_single_TEST) list ->
    Model_Def.variant -> (Model_Pattern.single * Model_Def.i_model_single_TEST) list
  val get_equal_descr: Model_Def.i_model_TEST -> Model_Pattern.single ->
    (Model_Pattern.single * Model_Def.i_model_single_TEST) list
  val unchecked_OLD_to_TEST: term list -> (term * Position.T) list

\<^isac_test>\<open>
(**)

\<close>
end

(**)
structure Pre_Conds(**) : PRE_CONDITIONS(**) =
struct
(**)
open Model_Def

type T = Pre_Conds_Def.T;  (*= (bool * term) list;*)
(*type checked = Pre_Conds_Def.T;*)
type unchecked = term list
type unchecked_TEST = (term * Position.T) list
type checked = bool * (bool * term) list
type checked_TEST = bool * (bool * (term * Position.T)) list

type env_subst = Env.T (*eg. [(fixes, r)]  .. with type adaptation, see switch_type*)
type env_eval = Env.T  (*eg. [(r, 7)    ]*)
type input = TermC.as_string list;

fun to_str ctxt (b, t) = pair2str (bool2str b, UnparseC.term ctxt t);
fun to_string ctxt pres = strs2str' (map (linefeed o (to_str ctxt)) pres);

fun eval _ _ (false, where_) = (false, where_) (*NOT ALL Free's have been substituted*)
  | eval ctxt where_rls (true, where_) =
    if Rewrite.eval_true ctxt [where_] where_rls
    then (true, where_)
    else (false, where_);

(*/---------------- old code before I_Model.T_TEST -------------------------------------------\*)
(* find most frequent variant v in itms *)
(* ATTENTION
 * spurios old code
 * misses variants with equal number of items, etc
*)
fun variants itms = ((distinct op =) o flat o (map (fn (_, vrs, _, _, _) => vrs))) itms;

(*version without typing problem*)
fun cnt itms v = (v, (length o (filter (curry op = v)) o flat o (map (fn (_, vrs, _, _, _) => vrs))) itms);
fun count_variants vts itms = map (cnt itms) vts;

fun max2 [] = raise ERROR "max2 of []"
  | max2 (y :: ys) =
    let
      fun mx (a, x) [] = (a, x)
  	    | mx (a, x) ((b, y) :: ys) = if x < y then mx (b, y) ys else mx (a, x) ys;
    in mx y ys end;

(* get the constant value from a penv *)
fun getval (id, values) = 
  case values of
	  [] => raise ERROR ("penv_value: no values in '" ^ UnparseC.term (ContextC.for_ERROR ()) id)
  | [v] => (id, v)
  | (v1 :: v2 :: _) => (case v1 of 
	      Const (\<^const_name>\<open>Program.Arbfix\<close>, _) => (id, v2)
	    | _ => (id, v1));

(* find the variant with most items already input *)
fun max_variant itms = 
    let val vts = (count_variants (variants itms)) itms;
    in if vts = [] then 0 else (fst o max2) vts end;

fun mk_e (Cor (_, iv)) = [getval iv]
  | mk_e (Syn _) = []
  | mk_e (Typ _) = [] 
  | mk_e (Inc (_, iv)) = [getval iv]
  | mk_e (Sup _) = []
  | mk_e (Mis _) = []
  | mk_e  _ = raise ERROR "mk_e: uncovered case in fun.def.";
fun mk_en vt (_, vts, _, _, itm_) = if member op = vts vt then mk_e itm_ else [];

(* extract the environment from an item list; takes the variant with most items *)
fun environment itms = 
  let val vt = max_variant itms
  in (flat o (map (mk_en vt))) itms end;(*\---------------- old code before I_Model.T_TEST -------------------------------------------/*)


(* feedback_TEST_to_string is a double of I_Model.* for mk_env *)
fun pen2str ctxt (t, ts) =
  pair2str (UnparseC.term ctxt t, (strs2str' o map (UnparseC.term  ctxt)) ts);
fun feedback_to_string_TEST ctxt (Cor_TEST ((d, ts), penv)) = 
    "Cor_TEST " ^ UnparseC.term ctxt (Input_Descript.join (d, ts)) ^ " ," ^ pen2str ctxt penv
  | feedback_to_string_TEST ctxt (Inp_TEST (descriptor, format)) =
    "Inp_TEST " ^ UnparseC.term ctxt descriptor ^ " " ^ format
  | feedback_to_string_TEST _ (Syn_TEST c) =
    "Syn_TEST " ^ c
  | feedback_to_string_TEST ctxt (Inc_TEST ((d, ts), penv)) = 
    "Inc_TEST " ^ UnparseC.term ctxt (Input_Descript.join (d, ts)) ^ " ," ^ pen2str ctxt penv
  | feedback_to_string_TEST ctxt (Sup_TEST (d, ts)) = 
    "Sup_TEST " ^ UnparseC.term ctxt (Input_Descript.join (d, ts))
fun feedback_TEST_to_string t = feedback_to_string_TEST (ContextC.for_ERROR ()) t;


(** check_variant_envs for PIDE **)

(*val all_atoms: term list -> bool*)
fun all_atoms ts =
  fold (curry and_) (map (fn t => TermC.is_atom (TermC.lhs t)) ts) true
fun switch_type (Free (id, _)) T = Free (id, T)
  | switch_type t _ = raise TERM ("switch_type not for", [t])
(*
  this repair is spurious and indicates the need for considering a separate language for Model-ing.
  E.g. "fixes" is used with 2 different types "bool list" and "real" in the problem-definition
    problem pbl_opti_univ : "univariate_calculus/Optimisation" = ..
      Given: "Constants fixes"
      Where: "0 < fixes"
      ...
     (1) Given: Const ("Input_Descript.Constants", "bool list \<Rightarrow> nam") $ Free ("fixes", "bool list")
     (2) Where: Const ("Orderings.ord_class.less", "real \<Rightarrow> real \<Rightarrow> bool") $
           Const ("Groups.zero_class.zero", "real") $ Free ("fixes", "real")

  The intermediately simplistic solution is to build a pair of environments, type env_pair,

  The simplistic design works for the instantiation of the demo-example
    "Constants [r = (7::real)]"
  but it could be extended to instantiations like
      "Constants [a = (1::real), b = (2::real)]"
    with the same Where: "0 < fixes" in the problem-definition
    if indexing / deletion of found elements from the association list is considered.

  For the purpose of future pimprovements a pair of environments, type env_pair, is built;
  where as key of the "outer" environment Free ("fixes", "real" kept to future improvements
  (by indexing "fixes_1", "fixes_2" in case the are more than one element in 
  "Input_Descript.Constants", "bool list..)
*)
fun repair_env env = 
  map (fn (id_inst, val_) => (switch_type id_inst (type_of val_), val_)) env
(* make an Env.T for Prec_Conds.eval *)
(* TODO:
 *1* eliminate TermC.isalist2list as soon as all list-type-terms are [singe-typed terms]
 *2* simplify code as soon as Cor_TEST, Inc_TEST, Sup_TEST have same argument-type
*)
fun mk_env (Model_Def.Cor_TEST ((descr, [ts]), _)) =
    let
    (*val ts = TermC.isalist2list ts*)
      val ts = if TermC.is_list ts then TermC.isalist2list ts else [ts]
    in
      (*compare Model_Pattern.empty_for*)
      case type_of descr of
        (Type ("fun", [Type ("List.list", [Type ("HOL.bool", [])]), _])) (*SideConditions \<longrightarrow> (sideconds, [(u / 2) \<up> 2 + (2 / v) \<up> 2 = r \<up> 2])*)
          => if all_atoms ts then map (fn t => (TermC.lhs t, TermC.rhs t)) ts else []
      | (Type ("fun", [Type ("List.list", [Type ("Real.real", [])]), _])) (*AdditionalValues [u, v]*)
          => []
      | (Type ("HOL.bool", [])) (*Precision (\<epsilon> = 0.01)*)
          => if all_atoms ts then map (fn t => (TermC.lhs t, TermC.rhs t)) ts else []
      | _ (*Maximum A*) => [(hd ts, TermC.empty (*dummy rhs*))]
    end
  | mk_env (Model_Def.Inc_TEST ((descr, [ts]), _)) =
    let
    (*val ts = TermC.isalist2list ts*)
      val ts = if TermC.is_list ts then TermC.isalist2list ts else [ts]
    in
      (*compare Model_Pattern.empty_for*)
      case type_of descr of
        (Type ("fun", [Type ("List.list", [Type ("HOL.bool", [])]), _])) (*SideConditions \<longrightarrow> (sideconds, [(u / 2) \<up> 2 + (2 / v) \<up> 2 = r \<up> 2])*)
          => if all_atoms ts then map (fn t => (TermC.lhs t, TermC.rhs t)) ts else []
      | (Type ("fun", [Type ("List.list", [Type ("Real.real", [])]), _])) (*AdditionalValues [u, v]*)
          => []
      | (Type ("HOL.bool", [])) (*Precision (\<epsilon> = 0.01)*)
          => if all_atoms ts then map (fn t => (TermC.lhs t, TermC.rhs t)) ts else []
      | _ (*Maximum A*) => [(hd ts, TermC.empty (*dummy rhs*))]
   end
  | mk_env (Model_Def.Sup_TEST (descr, [ts])) =      
    let
    (*val ts = TermC.isalist2list ts*)
      val ts = if TermC.is_list ts then TermC.isalist2list ts else [ts]
    in
      (*compare Model_Pattern.empty_for*)
      case type_of descr of
        (Type ("fun", [Type ("List.list", [Type ("HOL.bool", [])]), _])) (*SideConditions \<longrightarrow> (sideconds, [(u / 2) \<up> 2 + (2 / v) \<up> 2 = r \<up> 2])*)
          => if all_atoms ts then map (fn t => (TermC.lhs t, TermC.rhs t)) ts else []
      | (Type ("fun", [Type ("List.list", [Type ("Real.real", [])]), _])) (*AdditionalValues [u, v]*)
          => []
      | (Type ("HOL.bool", [])) (*Precision (\<epsilon> = 0.01)*)
          => if all_atoms ts then map (fn t => (TermC.lhs t, TermC.rhs t)) ts else []
      | _ (*Maximum A*) => [(hd ts, TermC.empty (*dummy rhs*))]
    end
  | mk_env feedb =
    raise ERROR ("mk_env not reasonable for " ^  quote (feedback_TEST_to_string feedb))
fun mk_env_subst equal_descr_pairs =
  map (fn ((_, (_, id)), (_, _, _, _, (feedb, _)))
        => (mk_env feedb) |> map fst (*dummy rhs dropped*)|> map (pair id)) equal_descr_pairs
  |> flat
  |> repair_env
fun mk_env_eval equal_descr_pairs =
  map (fn (_, (_, _, _, _, (feedb, _))) => mk_env feedb) equal_descr_pairs
  |> flat
  |> filter_out (fn (_, b) => b = TermC.empty) (*drop pairs with dummy rhs*)
fun check_variant_envs ctxt where_rls pres env_subst env_eval =
  let
    val pres_subst = map (TermC.subst_atomic_all env_subst) pres;
    val full_subst = map (fn (bool, t) => (bool, subst_atomic env_eval t)) pres_subst
    val evals = map (eval ctxt where_rls) full_subst
  in
    (foldl and_ (true, map fst evals), pres_subst)
  end


(** check_variant_envs for PIDE **)

fun get_i_model equal_descr_pairs = map snd equal_descr_pairs
fun filter_variants pairs i = 
  filter (fn (_, (_, variants, _, _, _)) => member op= variants i) pairs
(*get descriptor of those items which can contribute to Subst.T and Env.T *)
(*val get_dscr: feedback_TEST -> descriptor option*)
fun get_dscr (Cor_TEST ((descr, _), _)) = SOME descr
  | get_dscr (Inc_TEST ((descr, _), _)) = SOME descr
  | get_dscr (Sup_TEST (descr, _)) = SOME descr
  | get_dscr _
    (*other feedback cannot contribute to I_Model.mk_env_TEST and Pre_Conds.mk_env_TEST*) = NONE
(*val get_descr: I_Model.single_TEST -> descriptor option*)
fun get_descr (_, _, _, _, (feedb, _)) =
  case get_dscr feedb of NONE => NONE | some_descr => some_descr
(*val get_equal_descr: I_Model.T_TEST -> Model_Pattern.single ->
    (Model_Pattern.single * I_Model.single_TEST) list*)
fun get_equal_descr i_model (m_patt_single as (_, (descr, _))) =
  let
    val equal_variants = 
      filter (fn i_single => case get_descr i_single of
          NONE => false (*--------vvvvv*)
        | SOME descr' => descr' = descr) (*probl_POS*) i_model
    in
      (map (pair m_patt_single) equal_variants)
    end
fun arrange_args2 [] _ _ _(*all have equal length*) = []
  | arrange_args2 (b :: bs) (c :: cs) (d :: ds) (cnt, all) =
    ((b, c, d), nth cnt all) :: (arrange_args2 bs cs ds (cnt + 1, all)) 
  | arrange_args2 _ _ _ _ = raise ERROR "I_Model.arrange_args2 UnequalLengths"
fun sep_variants_envs model_patt i_model =
  let
    val equal_descr_pairs =
      map (get_equal_descr i_model) model_patt
      |> flat
    val all_variants =
        map (fn (_, (_, variants, _, _, _)) => variants) equal_descr_pairs
        |> flat
        |> distinct op =
    val variants_separated = map (filter_variants equal_descr_pairs) all_variants
    val i_models = map get_i_model variants_separated
(*NEW*)  
    val restricted_to_given = map (filter (fn ((m_field, (_, _)), (_, _, _, _, (_, _))) =>
      m_field = "#Given")) variants_separated
    val envs_subst = map mk_env_subst restricted_to_given
    val envs_eval =  map mk_env_eval restricted_to_given
  in
    arrange_args2 i_models envs_subst envs_eval (1, all_variants)
  end
(* smash all environments into one; this works only for old test before Maximum-example *)
fun sep_variants_envs_OLD model_patt i_model =
  let
    val equal_descr_pairs =
      map (get_equal_descr i_model) model_patt
      |> flat
    val all_variants =
        map (fn (_, (_, variants, _, _, _)) => variants) equal_descr_pairs
        |> flat
        |> distinct op =
    val variants_separated = map (filter_variants equal_descr_pairs) all_variants
    val restricted_to_given = map (filter (fn ((m_field, (_, _)), (_, _, _, _, (_, _))) =>
      m_field = "#Given")) variants_separated
    val envs_subst = map mk_env_subst restricted_to_given
    val envs_eval =  map mk_env_eval restricted_to_given
  in
    (flat envs_subst, flat envs_eval)
  end

(* expects the precondition from Problem, ie. needs substitution *)
fun check _ _ [] _ _ = (true, [])
  | check ctxt where_rls pres pbl (_ (*I_Model.variants UNNECESSARY: I_Model.T filtered for*)) =
    let
      val env = environment pbl;
      val pres' = map (TermC.subst_atomic_all env) pres;
      val evals = map (eval ctxt where_rls) pres';
    in (foldl and_ (true, map fst evals), evals) end;
(* expects the precondition from PIDE, ie. already substituted *)
fun check_TEST _ _ [] _ = (true, [])
  | check_TEST ctxt where_rls preconds env_eval =
    let
      val full_subst =
        map (fn (t, pos) => (subst_atomic env_eval t, pos)) preconds
      val evals =
        map (fn (term, pos) => (Rewrite.eval_true ctxt [term] where_rls, (term, pos))) full_subst
    in
      (foldl and_ (true, map fst evals), (map fst evals) ~~ preconds)
    end;
(* expects the precondition from Problem, ie. needs substitution *)
fun check_OLD _ _ [] _  = (true, [])
  | check_OLD ctxt where_rls pres (model_patt, i_model) =
    let
      val (env_subst, env_eval) = sep_variants_envs_OLD model_patt i_model
      val pres_subst = map (TermC.subst_atomic_all env_subst) pres;
      val full_subst = map (fn (bool, t) => (bool, subst_atomic env_eval t)) pres_subst
      val evals = map (eval ctxt where_rls) full_subst
    in
      (foldl and_ (true, map fst evals), pres_subst)
    end;


fun unchecked_OLD_to_TEST pres = map (fn t => (t, Position.none)) pres

(**)end(**)
